@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2006 - 2017
+  Copyright (c) 2006 - 2018
   CLST - Radboud University
   ILK  - Tilburg University
 
@@ -28,17 +28,21 @@
 #define UCTO_TOKENIZE_H
 
 #include <vector>
+#include <set>
 #include <map>
 #include <sstream>
 #include <stdexcept>
-#include "ucto/unicode.h"
-#include "ucto/setting.h"
+#include "libfolia/folia.h"
 #include "ticcutils/LogStream.h"
+#include "ticcutils/Unicode.h"
+#include "ucto/setting.h"
 
 class TextCat;
 
 namespace Tokenizer {
 
+  using namespace icu;
+
   std::string Version();
   std::string VersionName();
 
@@ -51,8 +55,7 @@ namespace Tokenizer {
     BEGINQUOTE                  = 16,
     ENDQUOTE                    = 32,
     TEMPENDOFSENTENCE           = 64,
-    LISTITEM                    = 128, //reserved for future use
-    TITLE                       = 256 //reserved for future use
+    LINEBREAK                   = 128
   };
 
   std::ostream& operator<<( std::ostream&, const TokenRole& );
@@ -96,8 +99,10 @@ namespace Tokenizer {
   public:
     TokenizerClass();
     ~TokenizerClass();
-    bool init( const std::string& ); // init from a configfile
-    bool init( const std::vector<std::string>& ); // init 1 or more languages
+    bool init( const std::string&,
+	       const std::string& ="" ); // init from a configfile
+    bool init( const std::vector<std::string>&,
+	       const std::string& ="" ); // init 1 or more languages
     bool reset( const std::string& = "default" );
     void setErrorLog( TiCC::LogStream *os );
 
@@ -109,12 +114,16 @@ namespace Tokenizer {
 
     //Tokenize from input stream to a vector of Tokens
     // non greedy. Stops after the first full sentence is detected.
+    // NOTE: may return more then one sentence, when more sentences are present
+    // on 1 line!
     // should be called multiple times until EOF
     std::vector<Token> tokenizeStream( std::istream&,
 				       const std::string& = "default" );
 
     // Tokenize from an input stream to a UTF8 string (representing a sentence)
     // non greedy. Stops after the first full sentence is detected.
+    // NOTE: may return more then one sentence, when more sentences are present
+    // on 1 line!
     // should be called multiple times until EOF
     std::string tokenizeSentenceStream( std::istream&,
 					const std::string& = "default" );
@@ -149,6 +158,7 @@ namespace Tokenizer {
 
     //return the sentence with the specified index in a Token vector;
     std::vector<Token> getSentence( int );
+    void extractSentencesAndFlush( int, std::vector<Token>&, const std::string& );
 
     //Get all sentences as a vector of strings (UTF-8 encoded)
     std::vector<std::string> getSentences();
@@ -173,9 +183,9 @@ namespace Tokenizer {
     bool setUppercase( bool b=true ) { bool t = uppercase; uppercase = b; if (b) lowercase = false; return t; };
     bool getUppercase() const { return uppercase; }
 
-    //Enable sentence-bound detection
-    bool setSentenceDetection( bool b=true ) { bool t = detectBounds; detectBounds = b; return t; }
-    bool getSentenceDetection() const { return detectBounds; }
+    //Enable sentence splitting only
+    bool setSentenceSplit( bool b=true ) { bool t = splitOnly; splitOnly = b; return t; }
+    bool getSentenceSplit() const { return splitOnly; }
 
     //Enable paragraph detection
     bool setParagraphDetection( bool b=true ) { bool t = detectPar; detectPar = b; return t; }
@@ -185,6 +195,10 @@ namespace Tokenizer {
     bool setQuoteDetection( bool b=true ) { bool t = detectQuotes; detectQuotes = b; return t; }
     bool getQuoteDetection() const { return detectQuotes; }
 
+    //Enable language detection
+    bool setLangDetection( bool b=true ) { bool t = doDetectLang; doDetectLang = b; return t; }
+    bool getLangDetection() const { return doDetectLang; }
+
     //Enable filtering
     bool setFiltering( bool b=true ) {
       bool t = doFilter; doFilter = b; return t;
@@ -196,6 +210,8 @@ namespace Tokenizer {
     }
     bool getPunctFilter() const { return doPunctFilter; };
 
+    std::string setTextRedundancy( const std::string& );
+
     // set normalization mode
     std::string setNormalization( const std::string& s ) {
       return normalizer.setMode( s );
@@ -209,7 +225,7 @@ namespace Tokenizer {
     void setLanguage( const std::string& l ){ default_language = l; };
 
     // set eos marker
-    UnicodeString setEosMarker( const std::string& s = "<utt>") { UnicodeString t = eosmark; eosmark =  folia::UTF8ToUnicode(s); return t; };
+    UnicodeString setEosMarker( const std::string& s = "<utt>") { UnicodeString t = eosmark; eosmark = TiCC::UnicodeFromUTF8(s); return t; };
     UnicodeString getEosMarker( ) const { return eosmark; }
 
     bool setNormSet( const std::string& );
@@ -227,6 +243,7 @@ namespace Tokenizer {
     const std::string setTextClass( const std::string& cls) {
       std::string res = inputclass;
       inputclass = cls;
+      outputclass = cls;
       return res;
     }
     const std::string getInputClass( ) const { return inputclass; }
@@ -261,6 +278,9 @@ namespace Tokenizer {
 		       bool,
 		       const std::string&,
 		       const UnicodeString& ="" );
+    int tokenizeLine( const UnicodeString&,
+		      const std::string&,
+		      const std::string& );
 
     bool detectEos( size_t, const UnicodeString&, const Quoting& ) const;
     void detectSentenceBounds( const int offset,
@@ -276,7 +296,6 @@ namespace Tokenizer {
     bool u_isquote( UChar32,
 		    const Quoting& ) const;
     std::string checkBOM( std::istream& );
-    void outputTokensDoc( folia::Document&, const std::vector<Token>& ) const;
     void outputTokensDoc_init( folia::Document& ) const;
 
     int outputTokensXML( folia::FoliaElement *,
@@ -286,7 +305,7 @@ namespace Tokenizer {
     void tokenizeSentenceElement( folia::FoliaElement *,
 				  const std::string& );
 
-    UnicodeNormalizer normalizer;
+    TiCC::UnicodeNormalizer normalizer;
     std::string inputEncoding;
 
     UnicodeString eosmark;
@@ -303,9 +322,6 @@ namespace Tokenizer {
     //verbose tokenisation mode
     bool verbose;
 
-    //detect sentence bounds?
-    bool detectBounds;
-
     //detect quotes?
     bool detectQuotes;
 
@@ -315,12 +331,22 @@ namespace Tokenizer {
     //filter all punctuation characters (default off)
     bool doPunctFilter;
 
+    // only sentence spliiting?
+    bool splitOnly;
+
     //detect paragraphs?
     bool detectPar;
 
     //has a paragraph been signaled?
     bool paragraphsignal;
 
+    //has do we attempt to assign languages?
+    bool doDetectLang;
+
+    //has do we percolate text up from <w> to <s> and <p> nodes? (FoLiA)
+    // values should be: 'full', 'minimal' or 'none'
+    std::string text_redundancy;
+
     //one sentence per line output
     bool sentenceperlineoutput;
     bool sentenceperlineinput;
