@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2006 - 2017
+  Copyright (c) 2006 - 2018
   CLST - Radboud University
   ILK  - Tilburg University
 
@@ -37,17 +37,23 @@
 #include "ticcutils/FileUtils.h"
 #include "ticcutils/PrettyPrint.h"
 #include "ticcutils/LogStream.h"
+#include "ticcutils/Unicode.h"
 #include "libfolia/folia.h"
-#include "ucto/unicode.h"
 #include "ucto/setting.h"
 
 using namespace std;
-using namespace TiCC;
 
-#define LOG *Log(theErrLog)
+#define LOG *TiCC::Log(theErrLog)
+
+#ifndef UCTODATA_DIR
+#define UCTODATA_DIR string(SYSCONF_PATH) + "/ucto/"
+#endif
 
 namespace Tokenizer {
-  string defaultConfigDir = string(SYSCONF_PATH) + "/ucto/";
+
+  using namespace icu;
+
+  string defaultConfigDir = UCTODATA_DIR;
 
   enum ConfigMode { NONE, RULES, ABBREVIATIONS, ATTACHEDPREFIXES,
 		    ATTACHEDSUFFIXES, PREFIXES, SUFFIXES, TOKENS, UNITS,
@@ -112,12 +118,12 @@ namespace Tokenizer {
     uConfigError( const string& s, const string& f ):
       invalid_argument( "ucto: " + s + " (" + f + ")"  ){};
     uConfigError( const UnicodeString& us, const string& f ):
-      uConfigError( folia::UnicodeToUTF8(us), f ){};
+      uConfigError( TiCC::UnicodeToUTF8(us), f ){};
   };
 
   class uLogicError: public std::logic_error {
   public:
-    uLogicError( const string& s ): logic_error( "ucto: logic error:" + s ){};
+    explicit uLogicError( const string& s ): logic_error( "ucto: logic error:" + s ){};
   };
 
   ostream& operator<<( ostream& os, const Quoting& q ){
@@ -188,7 +194,7 @@ namespace Tokenizer {
 
   Rule::Rule( const UnicodeString& _id, const UnicodeString& _pattern):
     id(_id), pattern(_pattern) {
-    regexp = new UnicodeRegexMatcher( pattern, id );
+    regexp = new TiCC::UnicodeRegexMatcher( pattern, id );
   }
 
   ostream& operator<< (std::ostream& os, const Rule& r ){
@@ -230,12 +236,27 @@ namespace Tokenizer {
       delete rule;
     }
     rulesmap.clear();
-    delete theErrLog;
+  }
+
+  set<string> Setting::installed_languages() {
+    // we only return 'languages' which are installed as 'tokconfig-*'
+    //
+    vector<string> files = TiCC::searchFilesMatch( defaultConfigDir, "tokconfig-*" );
+    set<string> result;
+    for ( auto const& f : files ){
+      string base = TiCC::basename(f);
+      size_t pos = base.find("tokconfig-");
+      if ( pos == 0 ){
+	string lang = base.substr( 10 );
+	result.insert( lang );
+      }
+    }
+    return result;
   }
 
   bool Setting::readrules( const string& fname ){
     if ( tokDebug > 0 ){
-      *theErrLog << "%include " << fname << endl;
+      LOG << "%include " << fname << endl;
     }
     ifstream f( fname );
     if ( !f ){
@@ -243,12 +264,12 @@ namespace Tokenizer {
     }
     else {
       string rawline;
-      while ( getline(f,rawline) ){
-	UnicodeString line = folia::UTF8ToUnicode(rawline);
+      while ( getline( f, rawline ) ){
+	UnicodeString line = TiCC::UnicodeFromUTF8(rawline);
 	line.trim();
 	if ((line.length() > 0) && (line[0] != '#')) {
 	  if ( tokDebug >= 5 ){
-	    *theErrLog << "include line = " << rawline << endl;
+	    LOG << "include line = " << rawline << endl;
 	  }
 	  const int splitpoint = line.indexOf("=");
 	  if ( splitpoint < 0 ){
@@ -266,14 +287,14 @@ namespace Tokenizer {
 
   bool Setting::readfilters( const string& fname ){
     if ( tokDebug > 0 ){
-      *theErrLog << "%include " << fname << endl;
+      LOG << "%include " << fname << endl;
     }
     return filter.fill( fname );
   }
 
   bool Setting::readquotes( const string& fname ){
     if ( tokDebug > 0 ){
-      *theErrLog << "%include " << fname << endl;
+      LOG << "%include " << fname << endl;
     }
     ifstream f( fname );
     if ( !f ){
@@ -281,12 +302,12 @@ namespace Tokenizer {
     }
     else {
       string rawline;
-      while ( getline(f,rawline) ){
-	UnicodeString line = folia::UTF8ToUnicode(rawline);
+      while ( getline( f, rawline ) ){
+	UnicodeString line = TiCC::UnicodeFromUTF8(rawline);
 	line.trim();
 	if ((line.length() > 0) && (line[0] != '#')) {
 	  if ( tokDebug >= 5 ){
-	    *theErrLog << "include line = " << rawline << endl;
+	    LOG << "include line = " << rawline << endl;
 	  }
 	  int splitpoint = line.indexOf(" ");
 	  if ( splitpoint == -1 )
@@ -314,7 +335,7 @@ namespace Tokenizer {
 
   bool Setting::readeosmarkers( const string& fname ){
     if ( tokDebug > 0 ){
-      *theErrLog << "%include " << fname << endl;
+      LOG << "%include " << fname << endl;
     }
     ifstream f( fname );
     if ( !f ){
@@ -322,12 +343,12 @@ namespace Tokenizer {
     }
     else {
       string rawline;
-      while ( getline(f,rawline) ){
-	UnicodeString line = folia::UTF8ToUnicode(rawline);
+      while ( getline( f, rawline ) ){
+	UnicodeString line = TiCC::UnicodeFromUTF8(rawline);
 	line.trim();
 	if ((line.length() > 0) && (line[0] != '#')) {
 	  if ( tokDebug >= 5 ){
-	    *theErrLog << "include line = " << rawline << endl;
+	    LOG << "include line = " << rawline << endl;
 	  }
 	  if ( ( line.startsWith("\\u") && line.length() == 6 ) ||
 	       ( line.startsWith("\\U") && line.length() == 10 ) ){
@@ -343,10 +364,40 @@ namespace Tokenizer {
     return true;
   }
 
+  UnicodeString escape_regex( const UnicodeString& entry ){
+    UnicodeString result;
+    for ( int i=0; i < entry.length(); ++i ){
+      switch ( entry[i] ){
+      case '?':
+      case '^':
+      case '$':
+      case '[':
+      case ']':
+      case '(':
+      case ')':
+      case '{':
+      case '}':
+      case '*':
+      case '.':
+      case '+':
+      case '|':
+      case '-':
+	if ( i == 0 || entry[i-1] != '\\' ){
+	  // not escaped
+	  result += "\\";
+	}
+	// fallthrough
+      default:
+	result += entry[i];
+      }
+    }
+    return result;
+  }
+
   bool Setting::readabbreviations( const string& fname,
 				   UnicodeString& abbreviations ){
     if ( tokDebug > 0 ){
-      *theErrLog << "%include " << fname << endl;
+      LOG << "%include " << fname << endl;
     }
     ifstream f( fname );
     if ( !f ){
@@ -354,15 +405,17 @@ namespace Tokenizer {
     }
     else {
       string rawline;
-      while ( getline(f,rawline) ){
-	UnicodeString line = folia::UTF8ToUnicode(rawline);
+      while ( getline( f, rawline ) ){
+	UnicodeString line = TiCC::UnicodeFromUTF8(rawline);
 	line.trim();
 	if ((line.length() > 0) && (line[0] != '#')) {
 	  if ( tokDebug >= 5 ){
-	    *theErrLog << "include line = " << rawline << endl;
+	    LOG << "include line = " << rawline << endl;
 	  }
-	  if ( !abbreviations.isEmpty())
+	  line = escape_regex( line );
+	  if ( !abbreviations.isEmpty()){
 	    abbreviations += '|';
+	  }
 	  abbreviations += line;
 	}
       }
@@ -438,7 +491,7 @@ namespace Tokenizer {
 		 UnicodeString &line,
 		 const string& fn ){
     try {
-      UnicodeRegexMatcher m( "\\s+" );
+      TiCC::UnicodeRegexMatcher m( "\\s+" );
       vector<UnicodeString> usv;
       m.split( line, usv );
       for ( const auto& us : usv  ){
@@ -459,7 +512,7 @@ namespace Tokenizer {
 
   void split( const string& version, int& major, int& minor, string& sub ){
     vector<string> parts;
-    size_t num = split_at( version, parts, "." );
+    size_t num = TiCC::split_at( version, parts, "." );
     major = 0;
     minor = 0;
     sub.clear();
@@ -497,7 +550,8 @@ namespace Tokenizer {
   }
 
   bool Setting::read( const string& settings_name,
-		      int dbg, LogStream* ls ) {
+		      const string& add_tokens,
+		      int dbg, TiCC::LogStream* ls ) {
     tokDebug = dbg;
     theErrLog = ls;
     map<ConfigMode, UnicodeString> pattern = { { ABBREVIATIONS, "" },
@@ -512,6 +566,14 @@ namespace Tokenizer {
     vector<string> meta_rules;
     string conffile = get_filename( settings_name );
 
+    if ( !TiCC::isFile( conffile ) ){
+      LOG << "Unable to open configfile: " << conffile << endl;
+      return false;
+    }
+    if ( !add_tokens.empty() && !TiCC::isFile( add_tokens ) ){
+      LOG << "Unable to open additional tokens file: " << add_tokens << endl;
+      return false;
+    }
     ifstream f( conffile );
     if ( f ){
       ConfigMode mode = NONE;
@@ -521,7 +583,7 @@ namespace Tokenizer {
       }
       int rule_count = 0;
       string rawline;
-      while ( getline(f,rawline) ){
+      while ( getline( f, rawline ) ){
 	if ( rawline.find( "%include" ) != string::npos ){
 	  string file = rawline.substr( 9 );
 	  switch ( mode ){
@@ -572,7 +634,7 @@ namespace Tokenizer {
 	  continue;
 	}
 
-	UnicodeString line = folia::UTF8ToUnicode(rawline);
+	UnicodeString line = TiCC::UnicodeFromUTF8(rawline);
 	line.trim();
 	if ((line.length() > 0) && (line[0] != '#')) {
 	  if (line[0] == '[') {
@@ -599,7 +661,7 @@ namespace Tokenizer {
 			rule_count, line, set_file );
 	      break;
 	    case METARULES:
-	      meta_rules.push_back( folia::UnicodeToUTF8(line) );
+	      meta_rules.push_back( TiCC::UnicodeToUTF8(line) );
 	      break;
 	    case ABBREVIATIONS:
 	    case ATTACHEDPREFIXES:
@@ -652,7 +714,7 @@ namespace Tokenizer {
 	      break;
 	    case NONE: {
 	      vector<string> parts;
-	      split_at( rawline, parts, "=" );
+	      TiCC::split_at( rawline, parts, "=" );
 	      if ( parts.size() == 2 ) {
 		if ( parts[0] == "version" ){
 		  version = parts[1];
@@ -661,7 +723,7 @@ namespace Tokenizer {
 	    }
 	      break;
 	    default:
-	      throw uLogicError("unhandled case in switch");
+	      throw uLogicError( "unhandled case in switch" );
 	    }
 	  }
 	}
@@ -678,7 +740,20 @@ namespace Tokenizer {
 	quotes.add( "“„‟", "”" );
       }
 
-      string split = "%";
+      if ( !add_tokens.empty() ){
+	ifstream adt( add_tokens );
+	string line;
+	while ( getline( adt, line ) ){
+	  UnicodeString entry = TiCC::UnicodeFromUTF8(line);
+	  entry = escape_regex( entry );
+	  if ( !entry.isEmpty() ){
+	    if ( !pattern[TOKENS].isEmpty() ){
+	      pattern[TOKENS] += '|';
+	    }
+	    pattern[TOKENS] += entry;
+	  }
+	}
+      }
       // Create Rules for every pattern that is set
       // first the meta rules...
       for ( const auto& mr : meta_rules ){
@@ -688,6 +763,7 @@ namespace Tokenizer {
 			      set_file );
 	}
 	string nam = TiCC::trim( mr.substr( 0, pos ) );
+	string split = "%";
 	if ( nam == "SPLITTER" ){
 	  split = mr.substr( pos+1 );
 	  if ( split.empty() ) {
@@ -702,7 +778,7 @@ namespace Tokenizer {
 	  }
 	  continue;
 	}
-	UnicodeString name = folia::UTF8ToUnicode( nam );
+	UnicodeString name = TiCC::UnicodeFromUTF8( nam );
 	string rule = mr.substr( pos+1 );
 	if ( tokDebug > 5 ){
 	  LOG << "SPLIT using: '" << split << "'" << endl;
@@ -716,7 +792,7 @@ namespace Tokenizer {
 	vector<UnicodeString> undef_parts;
 	bool skip_rule = false;
 	for ( const auto& part : parts ){
-	  UnicodeString meta = folia::UTF8ToUnicode( part );
+	  UnicodeString meta = TiCC::UnicodeFromUTF8( part );
 	  ConfigMode mode = getMode( "[" + meta + "]" );
 	  switch ( mode ){
 	  case ORDINALS:
@@ -738,7 +814,7 @@ namespace Tokenizer {
 	    break;
 	  case NONE:
 	  default:
-	    new_parts.push_back( folia::UTF8ToUnicode(part) );
+	    new_parts.push_back( TiCC::UnicodeFromUTF8(part) );
 	    break;
 	  }
 	}
@@ -784,7 +860,11 @@ namespace Tokenizer {
       }
       LOG << "EOS markers: " << eosmarkers << endl;
       LOG << "Quotations: " << quotes << endl;
-      LOG << "Filter: " << filter << endl;
+      try {
+	LOG << "Filter: " << filter << endl;
+      }
+      catch (...){
+      }
     }
     return true;
   }
